static char rcsid[] = "@(#)$Id: sb_mem.c,v 2.3 2011-03-06 19:52:58 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "s_me.h"
#include "sb_imp.h"

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

S_(sb_init_buffer sb_init_mem)
static int sb_init_mem P_((struct stringbuffer *buffer));
static int sb_init_mem(buffer)
     struct stringbuffer *buffer;
{
    buffer->p->a.mem.string_count = 0;
    buffer->p->a.mem.strings      = NULL;
    return 1;
}

S_(sb_free_buffer sb_free_mem)
static void sb_free_mem P_((struct stringbuffer *buffer));
static void sb_free_mem(buffer)
     struct stringbuffer *buffer;
{
    int i;

    for (i = 0; i < buffer->p->a.mem.string_count; i++) {
	if (buffer->p->a.mem.strings[i].line)
	    free_string(&(buffer->p->a.mem.strings[i].line));
    }
    if (buffer->p->a.mem.strings)
	free(buffer->p->a.mem.strings);
    buffer->p->a.mem.strings      = NULL;
    buffer->p->a.mem.string_count = 0;
}

S_(sb_add_line_to_stringbuffer sb_add_line_to_mem)
static void sb_add_line_to_mem P_((struct stringbuffer *buffer,
				   const struct string *string,
				   int pg_flags));
static void sb_add_line_to_mem(buffer,string,pg_flags)
     struct stringbuffer *buffer;
     const struct string *string;
     int pg_flags;
{
    int ptr = buffer->p->a.mem.string_count;
    
    buffer->p->a.mem.strings = safe_realloc(buffer->p->a.mem.strings,
					    (buffer->p->a.mem.string_count+1) *
					    sizeof (buffer->p->a.mem.strings[0]));

    buffer->p->a.mem.strings[ptr].line     = dup_string(string);
    buffer->p->a.mem.strings[ptr].pg_flags = pg_flags;

    buffer->p->a.mem.string_count++;
}

S_(sb_linecount_stringbuffer sb_linecount_mem)
static int sb_linecount_mem P_((const struct stringbuffer *ptr));
static int sb_linecount_mem(ptr)
     const struct stringbuffer *ptr;
{
    return ptr->p->a.mem.string_count;
}

S_(sb_get_line_from_stringbuffer sb_get_line_from_mem)
static struct string *sb_get_line_from_mem P_((const struct 
					       stringbuffer *buffer,
					       int ptr, int *pg_flags));
static struct string *sb_get_line_from_mem(buffer,ptr,pg_flags)
     const struct stringbuffer *buffer;
     int ptr;
     int *pg_flags;
{
    if (ptr < 0 || ptr >= buffer->p->a.mem.string_count)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_mem",
	      "Bad index",0);

    *pg_flags = buffer->p->a.mem.strings[ptr].pg_flags;
    return dup_string(buffer->p->a.mem.strings[ptr].line);

}

struct sb_type   sb_in_mem = {
    sb_init_mem,
    sb_free_mem,
    sb_add_line_to_mem,
    sb_linecount_mem,
    sb_get_line_from_mem
};

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

