static char rcsid[] = "@(#)$Id: mailmsg2.c,v 2.18 2011-04-09 06:22:14 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.18 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  Based on Elm 2.4 src/mailmsg2.c. That code was following copyright:
 *
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************
 *  Incorparated Elm 2.5 code from src/sndmsg.c. 
 *  That code was following copyright:
 *
 *  The Elm Mail System
 *
 * This file and all associated files and documentation:
 *                      Copyright (c) 1988-1995 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** Interface to allow mail to be sent to users.  Part of ELM  **/

#include "def_elm.h"
#include "s_elm.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"mail");

#include "menu2.h"

static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static const unsigned char *cs2cus P_((const char *str));
static const unsigned char *cs2cus(str) 
     const char *str;
{
    return (const unsigned char *)str;
}

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

extern char version_buff[];

int  gotten_key;

static int verify_transmission P_((char *filename,
				   int already_has_text, 
				   struct copy_file *copy_file,
				   int force_cmd, int options, int *dsn, 
				   struct mailing_headers *headers,
				   struct Attachments *attachments,
				   charset_t file_set,
				   struct mailer_info  *mailer_info,
				   charset_t hdr_charset,
				   struct MailboxView *mailbox,
				   int encode_hdr,
				   struct AliasView *aview,
				   struct menu_context *page,
				   struct menu_context *prompt_area,
				   struct header_rec *parent_message));  

static void mail_sent P_((union any_fd fd, char * title, struct run_state *rs,
 			  int ret, int exit_stat));

static int append_sig P_((FILE *file,struct mailing_headers *headers,
			  struct menu_context *page));


static void mailing_message P_((int background));
static void mailing_message(background)
     int background;
{
    if (!background)
	lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmSendingMail,
			      "Sending mail..."));
    else
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmSendingMailBackground,
			  "Sending mail... in background"));
    
    FlushBuffer();
}

static const unsigned char * csUs P_((const char *str));
static const unsigned char * csUs(str)
     const char *str;
{
    return (const unsigned char *)str;
}

struct string * gen_From_buffer(current_header)
     const struct header_rec * current_header;
{
    struct string * From_buffer = NULL;

    struct addr_item *p;
    
    if (current_header->from) {

	int idx;
	int addr_item_count = 
	    addr_list_item_count(current_header->from);

	for (idx = 0; idx < addr_item_count; idx++) {

	    int group = -1;   /* Group phrases are not allowed on
				 From: -header field              */
	    const struct address * address = 
		addr_list_get_item(current_header->from,idx,&group);

	    const char          * addr     = address_get_ascii_addr(address);
	    const struct string * fullname = address_get_phrase(address);

	    if (From_buffer)
		add_ascii_to_string(From_buffer,s2us(", "));
	    else
		From_buffer = new_string(display_charset);

	    if (fullname && string_len(fullname)) {

		append_string(&From_buffer,fullname,1);

	    } else if (addr) {

		add_ascii_to_string(From_buffer,csUs(addr));

	    }
	}
    }

    return From_buffer;
}


enum mailer_capab get_mailer_level(mailer_info)
     struct mailer_info *mailer_info;
{
    int an = give_dt_enumerate_as_int(&allow_no_encoding);

    enum mailer_capab result = mailer_7bit;

    if (query_mailer_info(mailer_info,MI_HAVE_8BITMIME)) {
	DPRINT(Debug,8,(&Debug,"mailer supports 8BITMIME\n"));
	result        = mailer_8bit;
	
	if (query_mailer_info(mailer_info,MI_HAVE_BINARYMIME)) {
	    DPRINT(Debug,8,(&Debug,"mailer supports BINARYMIME\n"));
	    result        = mailer_binary;
	}
    }

    if (an >= 2) {

	/* Just send BINARY anyway */
	result     = mailer_binary;

	DPRINT(Debug,8,(&Debug,"just send binary anyway\n"));

    }  else if (an >= 1 &&
		result < mailer_8bit) {
	
	/* Just send 8BIT anyway */
	result     = mailer_8bit;

	DPRINT(Debug,8,(&Debug,"just 8 bit anyway\n"));
    }


    DPRINT(Debug,8,(&Debug,"get_mailer_level=%d\n",result));

    return result;
	
}

static void handle_mailer_options  P_((struct mailer_info *mailer_info,
				       mime_send_t *mime_info,
				       int *dsn,
				       struct Attachments *attachments,
				       int *reask));

static void handle_mailer_options(mailer_info,mime_info,dsn, attachments,
				  reask)
     struct mailer_info *mailer_info;
     mime_send_t *mime_info;
     int *dsn;
     struct Attachments *attachments;
     int *reask;  /* on reask loop */
{
    int i;
    int changed  = 0;

    if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
	DPRINT(Debug,8,(&Debug,"mailer supports DSN\n"));
	if(DSN_success && !reask) {
	    (*dsn) |= DSN_SUCCESS|DSN_FAILURE|DSN_DELAY;
	    DPRINT(Debug,8,(&Debug," dsn-success is set .. enabling\n"));
	}
    } else if (*dsn) {
	*dsn = 0;
	
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoDSN,
			  "Mailer do not support DSN, disabling it"));

	if (reask)
	    *reask = 1;

    }

    mime_info->raw_level    = get_mailer_level(mailer_info);
    
    for (i = 0; i < attachments->attachment_count; i++) {
	mime_t *part = & (attachments->attach_files[i]);
	
	if (ENCODING_BINARY == part->encoding &&
	    mime_info->raw_level < mailer_binary) {
	    part->encoding = ENCODING_BASE64;
	    changed++;
	}
		
	if (ENCODING_8BIT == part->encoding &&
	    mime_info->raw_level < mailer_8bit) {
	    part->encoding = ENCODING_BASE64;
	    changed++;
	}
    }

    if (changed) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmChangedAttachEnc,
			  "Changed encoding of %d attachments"),
		  changed);
	
	if (reask)
	    *reask = 1;
    }
}

static char * mmsg_copy_part P_((FILE * mailbox_file,mime_t *mime_rec,
			       FILE *outfile,int quote_l, int *need_enc));

static char * mmsg_copy_part(mailbox_file,mime_rec,outfile,quote_l,
			     need_enc)
     FILE * mailbox_file;
     mime_t *mime_rec;
     FILE *outfile;   /* MAY be NULL */
     int quote_l;
     int *need_enc;
{
    FILE *F;
    static int PART_COUNT = 0;
    struct out_state * state_out = NULL;
    struct in_state * state_in = new_in_state(STATE_in_file);
    char *fname = NULL;
    char *tmp;

    if (0 != fseek(mailbox_file,mime_rec->offset,SEEK_SET)) {
	DPRINT(Debug,3,(&Debug,
			"mmsg_copy_part: Seek failed to %ld\n",
			mime_rec->offset));
	
	goto FAIL;
    }

    tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir");
    if (!tmp)
	goto FAIL;

    state_out = new_out_state(STATE_out_file);

    fname = elm_message(FRM("%smsgpart.%d-%d"), 
			tmp, getpid (),
			PART_COUNT++);

    F = safeopen_rdwr(fname);
    if (!F) {
	DPRINT(Debug,3,(&Debug,"Can't create: %s\n",
			fname));

	lib_error(CATGETS(elm_msg_cat, MeSet, MeDecodeFailedCreate,
			  "Failed to create file for decoding."));

	goto FAIL;
    }
    
    set_out_state_file(F,state_out);
    set_in_state_file(mailbox_file,state_in);

    if (run_cte_decoder(mime_rec,state_in,state_out, &NULL_decode_opt)) {

	if (outfile) {
	    int ch;
	    int nl = 1;
	    	    
	    rewind(F);

	    while (EOF != (ch  = getc(F))) {

		if (nl && '[' == ch && quote_l)
		    putc('[',outfile);
		
		putc(ch,outfile);
		nl = '\n' == ch;
	    }
	}

    } else {

	/* FAILURE */
	unlink(fname);
	free(fname);
	fname = NULL;
    }
 
    free_out_state(&state_out);
    
    if (need_enc) {
	rewind(F);
	*need_enc = needs_encoding(F,NULL,0);

    }


    fclose(F);


    if (0) {
 FAIL:
	if (fname)	
	    free(fname);
	fname = NULL;	
    }
    free_in_state(&state_in);

    return fname;
}


static int convert_attach_message P_((FILE *infile,
				      struct header_rec *current_header,
				      enum mailer_capab  mailer_level,
				      FILE * outfile,
				      enum encoding      * encoding
				      ));
static int convert_attach_message(infile,current_header,mailer_level,outfile,
				  encoding)
     FILE *infile;
     struct header_rec *current_header;
     enum mailer_capab  mailer_level;
     FILE                   * outfile;
     enum encoding          * encoding;
{
    int r;
    struct out_state  * target = new_out_state(STATE_out_file);

    set_out_state_file(outfile,target);

    r = convert_remail_message_1(infile,current_header,mailer_level,
				 target,encoding,NULL);

    /* Does not close file */
    free_out_state(&target);

    DPRINT(Debug,8,(&Debug, 
		    "convert_attach_message=%d\n",
		    r)); 

    return r;
}

int attach_rfc822(attachments,hdr,mailbox_file,title,
		  mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     struct string      *title;
     enum mailer_capab  mailer_level;
{
    int ret = 0;
    mime_t attach;
    static int count = 0;
    char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir");
    char * filename;
    FILE * tmpfp;
    enum encoding encoding = ENCODING_7BIT;

    if (!tmp)
	return 0;
    
    filename = elm_message(FRM("%selmmsg.%d-%d"), 
			   tmp, getpid (), count++);
    
    tmpfp = safeopen_rdwr(filename);
    
    if (!tmpfp) {
	free(filename);
	return 0;
    }
     
    if (hdr->status  & MIME_MESSAGE) {

	enum encoding encoding1 = ENCODING_NONE;

	if (0 != mime_warnings(hdr)) {
	    
	    DPRINT(Debug,4,(&Debug, 
			    "attach_rfc822: Bad-MIME message .. remailing as is\n"));
	    goto as_is;
	}

	DPRINT(Debug,4,(&Debug, 
			"attach_rfc822: Doing possible mime conversions\n"));
	
	/* Do possible conversions   */

	if (!convert_attach_message(mailbox_file,hdr,
				    mailer_level,tmpfp,
				    &encoding1)) {

	    unlink(filename);
	    free(filename);
	    fclose(tmpfp);
	    
	    return 0;
	}

	/* Pick up the encoding from the message. */
	update_encoding (&encoding, encoding1);

    } else {
   		
	DPRINT(Debug,4,(&Debug, 
			"attach_rfc822: Non-MIME message .. remailing as is\n"));
    as_is:
       
	if (!copy_message_f(mailbox_file,hdr,tmpfp,
			    CM_REMOVE_ENVELOPE,NULL,
			    NULL)) {
	    
	    unlink(filename);
	    free(filename);
	    fclose(tmpfp);
	    
	    return 0;
	}

	/* Pick up the encoding from the message. */
	update_encoding (&encoding,
			 hdr->mime_rec.encoding);
    }


    mime_t_zero(&attach);
    attach.pathname0 = filename;
    attach.dispname  = new_string2(system_charset,s2us(filename));
		    
    attach.length = fsize(tmpfp);
    fclose (tmpfp);
    
    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = give_media_type2(MIME_TYPE_MESSAGE,"rfc822", 0);
    if (!attach.TYPE)
	mime_panic(__FILE__,__LINE__,"attach_rfc822",
		   "message/rfc822 is not known");
 
    if (title)
	attach.description = dup_string(title);

    attach.encoding = encoding;
    add_Attachments(attachments,&attach);

    return 1;
}

int attach_body(attachments,hdr,mailbox_file,mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     enum mailer_capab  mailer_level;
{

    int need_enc = 0;
    mime_t attach;
    
    mime_t_zero(&attach);

    attach.pathname0 = mmsg_copy_part(mailbox_file,& (hdr->mime_rec),
				      NULL,0,&need_enc);
    if (!attach.pathname0) {
	DPRINT(Debug,4, (&Debug, 
			 "Failed to copy messge \n"));
	return 0;
    }

    attach.dispname = new_string2(display_charset,
				  s2us(attach.pathname0));
    
    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = hdr->mime_rec.TYPE;
    attach.encoding     = hdr->mime_rec.encoding;
    attach.disposition  = hdr->mime_rec.disposition;
    
    if (attach.encoding >= ENCODING_NONE &&
	attach.encoding <= ENCODING_BINARY) {
	enum encoding  old = attach.encoding;

	if (need_enc & HAVE_BINARY) {
	    if (mailer_level >= mailer_binary)
		attach.encoding = ENCODING_BINARY;
	    else
		/* no -BBINARYMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	else if (need_enc & HAVE_8BIT) {
	    if (mailer_level >= mailer_8bit)
		/* Just send 8BIT anyway */
		attach.encoding = ENCODING_8BIT;
	    else
		/* no -B8BITMIME option */
		attach.encoding = ENCODING_QUOTED;
	}

	DPRINT(Debug,12,(&Debug,
			 "attach_body: %s need_enc=%d%s%s fixup encoding %d (%s) => %d (%s)\n",
			 attach.pathname0,
			 need_enc,
			 (need_enc & HAVE_BINARY) ? " BINARY" : "",
			 (need_enc & HAVE_8BIT)   ? " 8BIT"   : "",
			 old,ENCODING(old),
			 attach.encoding, ENCODING(attach.encoding)));	
    }



    if (hdr->mime_rec.description) {
	attach.description = 
	    dup_string(hdr->mime_rec.description);
    }
    
    if (hdr->mime_rec.TYPE_opts) {
	attach.TYPE_opts = 
	    copy_mime_param(hdr->mime_rec.TYPE_opts);
    }
    
    if (hdr->mime_rec.DISPOSITION_opts) {
	attach.DISPOSITION_opts = 
	    copy_mime_param(hdr->mime_rec.DISPOSITION_opts);
    }
    
    add_Attachments(attachments,&attach);
    
    return 1;
}

static int attach_part P_((struct Attachments *attachments,
			   struct header_rec  *hdr,
			   FILE               *mailbox_file,
			   int x,
			   enum mailer_capab  mailer_level));
static int attach_part(attachments,hdr,mailbox_file,x,mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     int x;
     enum mailer_capab  mailer_level;
{    
    mime_t *z = NULL;
    mime_t attach;
    int need_enc = 0;

    z = mime_parser_index(hdr->mime_rec.parser_data,x);
    if (!z) {
	DPRINT(Debug,4, (&Debug, "Failed to get part %d\n",x));
	return 0;
    }

    if (!z->TYPE) {
	DPRINT(Debug,4, (&Debug, "No type on part %d\n",x));
	return 0;
    }

    mime_t_zero(&attach);

    attach.pathname0 = mmsg_copy_part(mailbox_file,z,NULL,0,
				      &need_enc);
    if (!attach.pathname0) {
	DPRINT(Debug,4, (&Debug, 
			 "Failed to copy part %d/%d\n",x));
	return 0;
    }

    attach.dispname = new_string2(display_charset,s2us(attach.pathname0));

    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = z->TYPE;
    attach.encoding     = z->encoding;
    attach.disposition  = z->disposition;

    if (attach.encoding >= ENCODING_NONE &&
	attach.encoding <= ENCODING_BINARY) {
	enum encoding  old = attach.encoding;
	
	if (need_enc & HAVE_BINARY) {
	    if (mailer_level >= mailer_binary)
		attach.encoding = ENCODING_BINARY;
	    else
		/* no -BBINARYMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	else if (need_enc & HAVE_8BIT) {
	    if (mailer_level >= mailer_8bit)
		/* Just send 8BIT anyway */
		attach.encoding = ENCODING_8BIT;
	    else
		/* no -B8BITMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	
	DPRINT(Debug,12,(&Debug,
			 "attach_part: %s need_enc=%d%s%s fixup encoding %d (%s) => %d (%s)\n",
			 attach.pathname0,
			 need_enc,
			 (need_enc & HAVE_BINARY) ? " BINARY" : "",
			 (need_enc & HAVE_8BIT)   ? " 8BIT"   : "",
			 old,ENCODING(old),
			 attach.encoding, ENCODING(attach.encoding)));	
    }
			
    if (z->description) {
	attach.description = dup_string(z->description);
    }

    if (z->TYPE_opts) {
	attach.TYPE_opts = copy_mime_param(z->TYPE_opts);
    }

    if (z->DISPOSITION_opts) {
	attach.DISPOSITION_opts = copy_mime_param(z->DISPOSITION_opts);
    }

    add_Attachments(attachments,&attach);

    return 1;
}

/* Returns 0 if reading of body failed */
int mail(mail_index,options, headers,mailer_info,
	  mailbox,aview, parent_page,
	  selection_view,selection_idx,
	  body)
     int mail_index;
     int options;
     struct mailing_headers * headers;
     struct mailer_info *mailer_info;     
     struct MailboxView *mailbox /* for mail_index */;
     struct AliasView *aview;
     struct menu_context  *parent_page;
     struct MailboxView *selection_view;
     int selection_idx;
     struct text_block *body;
{
    int replying      = 0 != (options & MAIL_REPLYING);
    int copy_msg      = 0 != (options & MAIL_COPY_MSG);
    int edit_message  = 0 != (options & MAIL_EDIT_MSG);
    int forwarding    = 0 != (options & MAIL_FORWARDING);
    int mime_attach   = 0 != (options & MAIL_MIME_ATTCH);
    int copy_selection = 0 != (options & MAIL_COPY_SELECTION); 
    int delete_cancel  = 0 != (options & MAIL_DELETE_CANCEL);

    struct Attachments attachments = NULL_Attachments;

    /** Given the addresses and various other miscellany (specifically, 
	'copy-msg' indicates whether a copy of the current message should 
	be included, 'edit_message' indicates whether the message should 
	be edited) this routine will invoke an editor for the user and 
	then actually mail off the message. 

	In redraw is needed use 
	       use menu_trigger_redraw(parent_page)
    **/

    FILE *reply = NULL;
    char fname[SLEN], very_long_buffer[VERY_LONG_STRING];
	  
    int ch,  line_len;

    int      already_has_text = FALSE;		/* we need an ADDRESS */
    int	 signature_done = FALSE;
    int	 err;
    int reask_verify = FALSE;
    int dsn = 0;
    mime_send_t MIME_info;
    
    FILE * converted_buffer = NULL;      /* Temp file which hold converted
					    body parts -- now include mime
					    headers however ...
					 */
    static int conv_count = 0;
    char * temp1 = NULL;

    char title[80];
    int code = 0;
    struct copy_file COPY_FILE;

    charset_t   cur_editcharset = NULL;
    
    struct menu_context *page = NULL;
    struct menu_context *prompt_area = NULL;
    char *tmp;

    int from_addr_len;
    int ret = 0;

    tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir");
    if (!tmp)
	return 0;

    /* Fail if no mailer available */
    if (!mailer_info)
	return 0;

    /* Initialize structure */
    clear_mime_send_info(&MIME_info);

    /* free_mailing_headers will free this */
    if (!headers->env_from)
	headers->env_from = mailer_get_env_from(mailer_info);

    zero_copy_file(&COPY_FILE);

    dump_expanded_address(4,"Mailing to",headers->to);
    DPRINT(Debug,4, (&Debug,"   (with%s editing)\n",
		     edit_message? "" : "out"));
    
    gotten_key = 0;		/* ignore previously gotten encryption key */
    
    /** first generate the temporary filename **/
    

    elm_sfprintf(cur_editfile,sizeof cur_editfile,
		 FRM("%s%s%d"), 
		 tmp, temp_file, getpid());
    

    /* verify_transmission() displays new screen,
       we create context for it on here
    */
       
    page = new_menu_context();
    if (OPMODE_IS_INTERACTIVE(opmode)) {
	int LINES, COLUMNS;

	menu_ClearScreen(page);

	menu_get_sizes(page,&LINES, &COLUMNS);

	prompt_area = 
	    new_menu_subpage(page,LINES-4,4,
			     subpage_simple_noredraw,NULL);

    }

    /** if we're not retransmitting, create the file.. **/

    if ((reply = safeopen_rdwr(cur_editfile)) == NULL) {
	err = errno;
	DPRINT(Debug,1,(&Debug,  
			"Attempt to write to temp file %s failed with error %s (mail)\n",
			cur_editfile, strerror(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotCreateFile,
			  "Could not create file %s (%s)."),
		  cur_editfile, strerror(errno));
	
	goto fail_label;
	
    }
   
    (void) elm_chown(cur_editfile, userid, groupid);

    if (body) {
	if (! add_block_to_mail(body,
				&already_has_text,
				reply,
				& cur_editcharset)) {

	    DPRINT(Debug,1,(&Debug,  
			    "Failed to add body block to editfile\n"));
	    
	    goto fail_label;
	}
    }

    handle_mailer_options(mailer_info,&MIME_info,&dsn, 
			  &attachments, NULL);

    if ((mailbox && mail_index >= 0 &&
	 (copy_msg || forwarding)) 
	|| 
	copy_selection) {
	struct header_rec *  hdr = NULL;
	FILE              *  mailbox_file = NULL;              
       
	struct MailboxView *  W  = mailbox;
	int                   Wi = mail_index;

	if (copy_selection) {
	    W  = selection_view;
	    Wi = selection_idx;
	}

	if (!give_message_data(W,Wi,
			       &hdr,&mailbox_file,NULL,
			       mime_parse_routine)) {
	    
	    DPRINT(Debug,4,(&Debug, 
			    "Failed to retrieve original mail for replying/forwarding/canceled mail\n"));
	    
	    fclose(reply); reply = NULL;

	    goto fail_label;
	}

	if (forwarding || mime_attach || copy_msg) { 

	    if (!cur_editcharset)
		cur_editcharset = display_charset;

	    if (!mime_attach) {
		if (edit_message && mailbox_file) {
		    int NOHDR = forwarding ? noheaderfwd : noheader;
		    int NOQUOTE = forwarding && !quote_forward; 
		    int FORW = forwarding;

		    copy_message_f(mailbox_file,hdr,
				   reply,
				   ( NOQUOTE ? 0 : CM_PREFIX ) |
				   ( NOHDR ? CM_REMOVE_HEADER : 0 ) |
				   CM_REMOVE_ENVELOPE | CM_DECODE |
				   ( FORW ? CM_FORWARDING : 0 ) |
				   ( mime_body_keywords ? CM_QUOTE_L : 0 ) |
				   CM_ATTRIBUTION
				   /* I think it is good idea to use CM_FILT_HDR
				    * even when we don't have forwarding ... -KEH
				    */
				   | CM_FILT_HDR, cur_editcharset,
				   NULL);
		    already_has_text = TRUE;	/* we just added it, right? */
		} else if (mailbox_file) {
		    int NOHDR = forwarding ? noheaderfwd : noheader;
		    int FORW = forwarding;
		
		    copy_message_f(mailbox_file,hdr,
				   reply,
				   ( NOHDR ? CM_REMOVE_HEADER : 0 ) |
				   CM_REMOVE_ENVELOPE |
				   ( FORW ? CM_FORWARDING : 0 ) |
				   ( mime_body_keywords ? CM_QUOTE_L : 0 ) |
				   CM_ATTRIBUTION
				   /* I added CM_DECODE to here -KEH */
				   | CM_DECODE, cur_editcharset,
				   NULL);
		}

	    } else {

		struct string * title = NULL;
		struct string * From_buffer = gen_From_buffer(hdr);	    
		
		if (forwarding) {
		    if (From_buffer)
			title = format_string(CATGETS(elm_msg_cat, ElmSet, 
						      ElmForwardedMesg1,
						      "Forwarded message from %S"), 
					      From_buffer);
		    else
			title =	format_string(CATGETS(elm_msg_cat, ElmSet, 
						      ElmForwardedMesg2,
						      "Forwarded message from %s"), 
					      hdr->env_from);
		}

		/* Use MESSAGE/RFC822 to forward messages. */
		
		if (! attach_rfc822(&attachments, hdr,mailbox_file,title,
				    MIME_info.raw_level))
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailFileForward,
				      "Failed to create file for forwarding"));
		
		if (title)
		    free_string(&title);

		if (From_buffer)
		    free_string(&From_buffer);	       	    
	    }

	} else  { 
	    /* make referenced message as edited message */
	    header_list_ptr all_headers,walk;
	    int Flags, major;
	    const char *subtype;
	    
	    DPRINT(Debug,4, (&Debug,
			     "Not forwarding, copying message normally or mime attaching, making message as edited -- copy_selection=%d\n",
			     copy_selection));

	    if (skip_envelope(hdr,mailbox_file) == -1) {
		DPRINT(Debug,4, (&Debug, "Can't skip envelope of message -- will not  pick message for sending\n"));
		goto fail0;
	    }

	    all_headers = file_read_headers(mailbox_file,RHL_CHECK_HEADER);

	    
	    for (walk = all_headers; walk; walk = walk -> next_header) {
		const char * hdr_name = give_header_name(walk->header_name);

		if (0 == strincmp(hdr_name,"Content-",8) ||
		    0 == istrcmp(hdr_name,"MIME-Version") ||
		    0 == istrcmp(hdr_name,"Received") ||
		    0 == istrcmp(hdr_name,"Sender"))
		    continue;

		if (!add_to_mailing_header(headers,walk->header_name,
					   walk->body,
					   !(hdr->status & NOHDRENCODING),
					   hdr->header_charset,1)) {
		    DPRINT(Debug,4, (&Debug, "Header %s not copied\n",
				     hdr_name));
		}
	    }

	    dump_expanded_address(4,"after copy -- from ",headers->from);
	    dump_expanded_address(4,"after copy -- to   ",headers->to);
	    dump_expanded_address(4,"after copy -- cc   ",headers->cc);


	    if (! hdr->mime_rec.TYPE) {
		DPRINT(Debug,4, (&Debug, "no mime type -- will not  pick message for sending\n"));
		goto fail1;		
	    }
	    
	    Flags   =  get_type_flags(hdr->mime_rec.TYPE);
	    major   =  get_major_type_code(hdr->mime_rec.TYPE);
	    subtype =  get_subtype_name(hdr->mime_rec.TYPE);
	    
	    /* 1) Includes only one text part -- put it to edit buffer */
	    if (MIME_TYPE_TEXT == major && 0 == istrcmp(subtype,"Plain") &&
		DISP_INLINE == hdr->mime_rec.disposition) {
		
		const char *cs = NULL;

		char * filename =
		    mmsg_copy_part(mailbox_file,& (hdr->mime_rec),
				   reply,mime_body_keywords,NULL);
		
		if (filename) {
		    unlink(filename);
		    free(filename);
		}

		cs = get_mime_param_compat(hdr->mime_rec.TYPE_opts,
					   "charset");
		if (!cs) {
		    DPRINT(Debug,4, (&Debug, "No charset parameter?\n"));
		    goto fail1;
		}

		    cur_editcharset = MIME_name_to_charset(cs,0);
		    if (!cur_editcharset) {
			DPRINT(Debug,4, (&Debug, "charset parameter %s unknown?\n",
					 cs));
			goto fail1;
		    }
		    
		    if (charset_superset_of(display_charset,cur_editcharset)) {
			const char *MIME_name_c = 
			    get_charset_MIME_name(cur_editcharset);
			const char *MIME_name_d = 
			    get_charset_MIME_name(display_charset);

			DPRINT(Debug,4, (&Debug, 
					 "charset parameter %s (%s) is subset of display charset %s\n",
					 MIME_name_c ? 
					 MIME_name_c : 
					 "<no MIME name>",
					 cs,
					 MIME_name_d ?
					 MIME_name_d :
					 "<no MIME name>"));
			cur_editcharset = display_charset;
		    }
		    
		    DPRINT(Debug,4, (&Debug, "Added whole mail to editing buffer\n"));

		    /* 2) multipart MIXED */
		} else if ((Flags & MIME_MIXED) &&
			   DISP_INLINE == hdr->mime_rec.disposition &&
			   hdr->mime_rec.parser_data) {
		    
		    int s = 0;
		    int c = mime_parser_subparts(hdr->mime_rec.parser_data);
		    mime_t *z = NULL;
		    int x;

		    /* Pick first parts to editing buffer */
		    if (c > 0 &&
			(z = mime_parser_index(hdr->mime_rec.parser_data,0)) &&
			z->TYPE) {
			int major           =  get_major_type_code(z->TYPE);
			const char *subtype =  get_subtype_name(z->TYPE);
			
			const char *cs = NULL;

			if (MIME_TYPE_TEXT == major && 0 == istrcmp(subtype,"Plain") &&
			    DISP_INLINE == z->disposition) {
			    
			    char * filename =
				mmsg_copy_part(mailbox_file,z,reply,mime_body_keywords,
					       NULL);
			    
			    if (filename) {
				unlink(filename);
				free(filename);
			    }
			    
			    cs = get_mime_param_compat(z->TYPE_opts,
						       "charset");
			    if (!cs) {
				DPRINT(Debug,4, (&Debug, "No charset parameter?\n"));
				goto fail2;
			    }
			    
			    cur_editcharset = MIME_name_to_charset(cs,0);
			    if (!cur_editcharset) {
				DPRINT(Debug,4, (&Debug, "charset parameter %s unknown?\n",
						 cs));
				goto fail2;
			    }
			    
			    if (charset_superset_of(display_charset,cur_editcharset)) {

				const char *MIME_name_c = 
				    get_charset_MIME_name(cur_editcharset);
				const char *MIME_name_d = 
				    get_charset_MIME_name(display_charset);


				DPRINT(Debug,4, (&Debug, 
						 "charset parameter %s (%s) is subset of display charset %s\n",
						 MIME_name_c ?
						 MIME_name_c :
						 "<no MIME name>",
						 cs,
						 MIME_name_d ?
						 MIME_name_d :
						 "<no MIME name>"));
				cur_editcharset = display_charset;
			    }
			    
			    DPRINT(Debug,4, (&Debug, 
					     "Added first part of mail to editing buffer\n"));

			fail2:
			    
			    /* Start attachments from second part */
			    s = 1;
			    
			}
			
		    }
		    
		    for (x = s; x < c; x++) {
			
			if (! attach_part(&attachments, hdr,mailbox_file,x,
					  MIME_info.raw_level)) {
			    DPRINT(Debug,4, (&Debug, 
					     "Failed to copy part %d/%d\n",
					     x,c));
			    continue;
			}


			DPRINT(Debug,4, (&Debug, 
					 "Added part %d/%d\n",
					 x,c));


		    }
		    
		 
		    /* 3) Add just whole mail as attachment */
		} else {

		    if (! attach_body(&attachments, hdr,mailbox_file,
				      MIME_info.raw_level)) {
			DPRINT(Debug,4, (&Debug, 
					 "Failed to copy messge \n"));
			goto fail1;
		    }
 

		    DPRINT(Debug,4, (&Debug, 
				     "Added message\n"));
		    
		}

		if (delete_cancel)
		    delete_current_cancel(selection_view,mailbox_file);
	              
	fail1:
		if (all_headers)
		    delete_headers(&all_headers);

	}
	    
    }
    
    
 fail0:

    if (!cur_editcharset)
	cur_editcharset = display_charset;


    if (attach_files.attachment_count > 0) {
	int i;

	for (i = 0; i < attach_files.attachment_count; i++) {
	    add_Attachments(& attachments,&attach_files.attach_files[i]);
	    /* add_Attachments zeros attach_files.attach_files[i],
	       so array can be fre'ed later 
	    */
	}
	free_Attachments(&attach_files); /* Actually only frees array */
    }

    /* append signature now if we are going to use an external editor */
    /* Don't worry about the remote/local determination too much */
    
    {
	char *     editor_val = give_dt_estr_as_str(&editor_e,"editor");

	if (already_has_text || 
	    (editor_val  &&
	     strcmp(editor_val,"builtin") != 0 && 
	     strcmp(editor_val,"none") != 0)) {
	    signature_done = TRUE;

	    already_has_text |= append_sig(reply, headers, page);
	}
    }

    if (reply != NULL) {
	(void) fclose(reply);	
	reply = NULL;
    }

    /** Edit the message **/

    /* calculate default save_file name */
    if (COPY_FILE.copy_file)
	clear_copy_file(&COPY_FILE);        /* signals to not save a copy */
    
    if(auto_copy_sent) {
	if(save_by_name || save_by_alias) {
	    if(force_name) {
		/* signals save by 'to' logname */
		COPY_FILE.copy_file = format_string(FRM("="));
	    } else {
		/* conditional save by 'to' logname */
		COPY_FILE.copy_file = format_string(FRM("=?"));
	    }
	} else {
	    /* signals save to sentmail */
	    COPY_FILE.copy_file = format_string(FRM("<"));
	}
    } 
    
    /* tell to verify_transmission() */
    if (attachments.attachment_count > 0) 
	options |= MAIL_HAVE_ATTACHMENTS;
#ifdef USE_PGP
    if (mailbox && mail_index >= 0) {
	struct header_rec *hdr =  give_header(mailbox,mail_index);
	if (hdr &&
	    hdr->pgp & PGP_MESSAGE)
	    options |= MAIL_HAVE_PGP_ENCODED;
    }
#endif

    /* Now .mailheaders is read before mail is composed! */
    if (!copy_selection)
	import_mailheaders(headers);

    do { /* So we can return here if check_for_multipart() fails
	  * - K E H <hurtta+elm@ozone.FMI.FI>    */

	code = 0;
	reask_verify = 0;
	
	/* ask the user to confirm transmission of the message */
	if (OPMODE_IS_INTERACTIVE(opmode)) {

	    struct header_rec *parent_message;
	fatal_label:

	    parent_message = NULL;

	    menu_set_default(page);

	    if (replying && mailbox && mail_index >= 0) 
		parent_message =  give_header(mailbox,mail_index);
	    	    
	    ch = (edit_message? 'e' : '\0');
	    code = verify_transmission(cur_editfile, 
				       already_has_text, &COPY_FILE, ch,
				       options, &dsn,
				       headers, &attachments,
				       cur_editcharset,
				       mailer_info,
				       MIME_info.hdr_charset,
				       mailbox,
				       MIME_info.encode_hdr,
				       aview,
				       page,
				       prompt_area,
				       parent_message);	    

	    DPRINT(Debug,8,(&Debug,
			    " ... virify_transmission result = %d, attachment_count=%d\n",
			    code,attachments.attachment_count));

	    /* Empty ... */
	    if (-2 == code && !attachments.attachment_count) {
		goto fail_label;
	    }
	}

	err = can_open(cur_editfile,"r+");
	if (err) {
	    DPRINT(Debug,1,(&Debug,  
			    "Attempt to open file %s for reading failed with error %s (mail) -- can_open\n",
			    cur_editfile, strerror(err)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotOpenReply,
			      "Could not open reply file (%s)."), 
		      strerror(err));

	    goto fail_label;
	}

	if ((reply = fopen(cur_editfile,"r+")) == NULL) {
	    err = errno;
	    DPRINT(Debug,1,(&Debug,  
			    "Attempt to open file %s for reading failed with error %s (mail)\n",
			    cur_editfile, strerror(err)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotOpenReply,
			      "Could not open reply file (%s)."), 
		      strerror(err));

	    goto fail_label;
	}

	/* Append signature if not done earlier, so that need_eccoding
	 * takes account of 8-bit data on signature
	 */
	
	if (!signature_done) {
	    /* Go to the end of the file! */
	    fseek (reply, 0, 2);
	    append_sig(reply, headers, page);
	    rewind(reply);
	    signature_done = TRUE;
	}

	if (code >= 0) { /* If not canceled */
	    int r = mailer_restarted(mailer_info);

	    if (r > 0)
		handle_mailer_options(mailer_info,&MIME_info,&dsn, 
				      &attachments, &reask_verify);

	    if (r < 0) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailerDisconnected,
				  "Mailer disconnected! Can't send message. Use f)orget !"));


		reask_verify = TRUE; /* Go to verify_transmission again. */
	    }
	}

	MIME_info.encoding_top  = ENCODING_7BIT;/* Encoding for Multipart/ */
	  	 
#if 0	  
	if (allow_no_hdrencoding && headers->subject)
	    MIME_info.need_enc |= check_8bit_string (headers->subject);
#endif
	
	if (check_for_multipart(reply, &MIME_info,mailer_info,
				cur_editcharset) < 0) { 
	    /* Error in [include ...] */
	    if (code < 0)
		goto fail_label;

	    if (OPMODE_IS_INTERACTIVE(opmode)) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFixInclude,
				  "Please fix [include ...] lines!"));
		edit_message = FALSE;
		reask_verify = TRUE; /* Go to verify_transmission again. */
		continue;
	    }
	}

	if (converted_buffer)
	    fclose(converted_buffer);
	temp1 = elm_message(FRM("%selmcv-%d-%d"),
			    tmp, getpid (),
			    conv_count++);
	converted_buffer = safeopen_rdwr(temp1);
	if (!converted_buffer) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmOpenFailedCopy,
			      "Temp file open failed to in copy"));
	    free(temp1);
	    temp1 = NULL;
	    if (code < 0)
		goto fail_label;
	    if (! OPMODE_IS_INTERACTIVE(opmode))
		leave(0);   
	    edit_message = FALSE;
	    reask_verify = TRUE; /* Go to verify_transmission again. */
	    continue;
	}
	DPRINT(Debug,2,(&Debug,   "mail: using temp file (%s)\n",
			temp1));
	unlink(temp1);
	free(temp1);
	temp1 = NULL;

	if (!convert_text(reply,converted_buffer, &MIME_info,
			  cur_editcharset,text_charset,
#ifdef USE_PGP
			  pgp_status,
#else
			  0,
#endif
			  &attachments,
			  mailer_info,
			  page,prompt_area)) {
	    if (code < 0)
		goto fail_label;
	    if (! OPMODE_IS_INTERACTIVE(opmode))
		leave(0);  
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedConvertMessage,
			      "Failed convert message, fix errors and retry or f)orget it"));
	    reask_verify = TRUE; /* Go to verify_transmission again. */
	    edit_message = FALSE;	    
	} else
	    MIME_info.msg_is_multipart = MIME_info.top_parts_count > 1;


	DPRINT(Debug,8,(&Debug,
			" ... top_parts_count=%d, msg_is_multipart=%d\n",
			MIME_info.top_parts_count,
			MIME_info.msg_is_multipart));


	/* End of check_for_multipart failure loop */
    } while (OPMODE_IS_INTERACTIVE(opmode) && reask_verify);
	
    if (MIME_info.msg_is_multipart) {
	(void) mime_generate_boundary (MIME_info.mime_boundary,
				       sizeof MIME_info.mime_boundary);
	mime_params_add_compat(&(MIME_info.TYPE_opts_top), 
			       "boundary",
			       MIME_info.mime_boundary);
    }

    if (headers->from.addrs &&
	(from_addr_len = addr_list_item_count(headers->from.addrs)) > 0) {

	DPRINT(Debug,6,(&Debug,
			"From address already specified by user (len %d)\n",
			from_addr_len));

	dump_expanded_address(6,"From address",headers->from);

	if (from_addr_len > 1 || add_sender_header) {
	    headers->sender = from_addr_literal(mailer_info);
	    DPRINT(Debug,6,(&Debug,
			    "setting Sender address: %s\n",
			    headers->sender));
	}

    } else if (!query_mailer_info(mailer_info,MI_DONT_ADD_FROM))
	make_from_addr(&(headers->from),mailer_info);
    
    if (headers->env_from && env_from_changed(headers->env_from) &&
	add_sender_header && !headers->sender) {
	    headers->sender = from_addr_literal(mailer_info);
	    DPRINT(Debug,6,(&Debug,
			    "setting Sender address: %s\n",
			    headers->sender));	    
    }

    if (code < 0 /* Forget */) {

	struct MailboxView *cm = give_canceled_mail();

	char * lbuf = NULL;
	struct out_state  * fp_copy = new_out_state(STATE_out_file);
	FILE *F = NULL;
	time_t now = time(NULL);

	if (cm)
	    F = add_canceled_mail(&lbuf,headers,cm,now,-1,NULL,
				  -1L);
				  
	if (!F) {
	    if (lbuf)
		free(lbuf);

	    lbuf = elm_message(FRM("%s/%s"), 
			       home, dead_letter);

	    F = safeopen_rdwr(lbuf);
	} else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmVfyMessageKept1,
			      "Message kept.  Can be restored at next m)ail command."));
	}
	
	DPRINT(Debug,4,(&Debug, "Writing canceled mail to %s\n",
			    lbuf));
	
	if (NULL != F) {
	    set_out_state_file(F,fp_copy);
	    write_header_info(fp_copy,headers,TRUE,
			      &MIME_info);

	    copy_message_across(&MIME_info,fp_copy,TRUE,converted_buffer);

	}
	free_out_state(&fp_copy);
	if (F)
	    fclose(F);

	if (lbuf) {

	    if (last_canceled_mail)
		free(last_canceled_mail);
	    last_canceled_mail = lbuf;	    
	}
    } else {
	int r;
	struct mail_send_state * mailer = NULL;
	/* Only do prompts on interactive mode */
	struct menu_context *p = 
	    OPMODE_IS_INTERACTIVE(opmode) ? prompt_area : NULL;


	rewind(converted_buffer);
	if (COPY_FILE.copy_file) /* i.e. if copy_file contains a name */
	    save_copy(headers,&COPY_FILE, &MIME_info, 
		      converted_buffer, 
		      page,aview,p);  /* Need access for all open 
				       mailboxes/folders */

	/** write all header information into whole_msg_file **/

	/** try to write headers to new temp file **/
	
	DPRINT(Debug,6,(&Debug,   "Composition file='%s'\n", 
		   cur_editfile));
	
	dump_expanded_address(2,"From",headers->from);
	dump_expanded_address(2,"To",headers->to);
	dump_expanded_address(2,"CC",headers->cc);
	dump_expanded_address(2,"Bcc",headers->bcc);

	{
	    char **argvt = argv_from_headers(headers);
	    mailer = mailer_init(argvt,dsn,sendmail_verbose,mailer_info,
				 headers->env_from);
	    free_argv(&argvt);
	}

	if (!mailer) {
	    edit_message = FALSE;
	    fclose(reply);
	    reply = NULL;		

	    if (OPMODE_IS_INTERACTIVE(opmode))
		goto fatal_label;
	    
	    goto fail_label;
	}
	
	/* Let state routines be aware of encoding */
	if (MIME_info.encoding_top == ENCODING_BINARY)
	    set_out_state_EOLN(get_mail_outfd(mailer),1);

	write_header_info(get_mail_outfd(mailer),headers,
			  FALSE, &MIME_info);

	rewind(converted_buffer);

	if (!copy_message_across(&MIME_info,get_mail_outfd(mailer),
				 FALSE,converted_buffer)) {
	    edit_message = FALSE;
	    fclose(reply);
	    reply = NULL;	

	    if (OPMODE_IS_INTERACTIVE(opmode))
		goto fatal_label;
	    
	    goto fail_label;
	}
	
	if (headers->to.addrs) {
	    int cnt;
	    
	    if ((cnt = addr_list_item_count(headers->to.addrs)) == 1) {
		int group;
		const struct address * X = 
		    addr_list_get_item(headers->to.addrs,0,&group);
		const char * addr = address_get_ascii_addr(X);
		
		elm_sfprintf(title, sizeof title,
			     CATGETS(elm_msg_cat, ElmSet, ElmMailTo,
				     "Mail to %.50s..."),
			     addr ? addr : "(someone)");
		
	    } else
		elm_sfprintf(title, sizeof title,
			     CATGETS(elm_msg_cat, ElmSet, ElmMailToXRecipients,
				     "Mail to %d recipients (To:)"),
			     cnt);


	} else
	    elm_sfprintf(title, sizeof title,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailToSomeone,
				 "Mail to someone..."));
	      			    	
	r = mail_backend2(&mailer,mail_sent, 
			  MIME_info.encoding_top, title,mailing_message);
	  		
 	DPRINT(Debug,4,(&Debug,"mail_backend2 returned %d\n",r));

 	/* 0   may be:    start failed
 	 *                 mailing contined on backend
 	 *
  	 *  1             command completed, perhaps with failure 
 	 */
	
 	ret = 1;


	/* NOTE: succees failure, backgrouded status currently ignored */

	/* mark the "replied" status of the message */
	if (replying && mailbox && mail_index >= 0 && ret) {

	    struct header_rec *hdr =  give_header(mailbox,mail_index);
	    if (hdr && 	0 != (hdr->status && REPLIED_TO)) {
		
		DPRINT(Debug,4,(&Debug,
				"Adding replied mark to current message\n"));
		
		hdr->status |= REPLIED_TO;
		hdr->status_chgd = TRUE;
	    } else {
		DPRINT(Debug,8,(&Debug,
				"Not adding replied mark  -- getting header failed\n"));
	    }

	} else {
	    DPRINT(Debug,8,(&Debug,"Not adding replied mark %s%s%s\n",
			    replying ? "" : " -- not replying ",
			    mailbox ? "" : " -- no mailbox",
			    ret ? "" : " -- no succeed"));
	}
    }

    fclose(reply);
    reply = NULL;
	
 fail_label:	

    if (delete_cancel)
	sync_canceled_mails(selection_view);

    free_Attachments(&attachments);

    if (converted_buffer)
	fclose(converted_buffer);
    converted_buffer = NULL;

    /* Unlink temp file now.
     * This is a precaution in case the message was encrypted.
     * I.e. even though this file is readable by the owner only,
     * encryption is supposed to hide things even from someone
     * with root privelges. The best we can do is not let this
     * file just hang after we're finished with it.
     */

    if (0 ==unlink (cur_editfile)) {
	DPRINT(Debug,8,(&Debug,"Removing %s\n",
			cur_editfile));
    } 

#ifdef USE_PGP
    pgp_status=0;
#endif
    
    free_mime_send_info(&MIME_info);
    clear_copy_file(&COPY_FILE);


    if (prompt_area)
	erase_menu_context(&prompt_area);


    if (page)
	erase_menu_context(&page);

    if (parent_page)
	menu_trigger_redraw(parent_page);

    DPRINT(Debug,8,(&Debug,"mail=%d\n",ret));

    return ret;
}

static void mail_sent (fd,title,rs,ret,exit_code)
     union any_fd fd; 
     char * title;
     struct run_state *rs;
     int ret; 
     int exit_code;
{    
  lower_prompt(title);

  if (ret < 0) 
    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailSignal,
		      "%.30s fail: Signal?"),
	      get_mailer_path(fd.mail_fd));
  else if (ret > 0) {
    if (rs->save_errno)
      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailErrno,
			"Failed: %.30s: %.40s"),
		get_mailer_path(fd.mail_fd),
		strerror(rs->save_errno));
    else if (exit_code) {
      lib_error(
		CATGETS(elm_msg_cat, ElmSet, 
			ElmMailerReturnedError,
			"mailer returned error status %d"), 
		exit_code);
    } else
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailSent, "Mail sent!"));
  } else {
      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLostErrno,
			"%.30s lost: %.40s"),
		get_mailer_path(fd.mail_fd),
		strerror(rs->save_errno));
  }
}

void mail_form(mail_index, address_list, subj, mailbox, aview, page)
     int mail_index;
     struct addr_list *address_list;
     char *subj;
     struct MailboxView *mailbox;
     struct AliasView *aview;
     struct menu_context  *page;
{
    struct mailing_headers headers;
    struct expanded_address A;
    struct mailer_info *mailer_info   = get_mailer_info();

    if (!mailer_info)
	return;
    
    zero_mailing_headers(&headers);
    zero_expanded_address(&A);

    headers.subject = new_string2(display_charset,s2us(subj));

    addr_to_expanded(&A,address_list,mailer_info,aview);
    copy_expanded_address(&headers.to, A);
    
    mail(mail_index, NO, &headers,mailer_info,
	 mailbox /* save_copy needs */,
	 aview, page,
	 NULL, -1, NULL);

    /* mail() does menu_trigger_redraw() do not need to reexamine */

    free_expanded_address(&A);
    free_mailing_headers(&headers);
    free_mailer_info(&mailer_info);

    return;
}


void dsn_menu P_((int *));
void dsn_menu(dsn)
     int * dsn;
{
  static int hdr_only = 0;
  static int failure  = 1;
  static int delay    = 1;
  static int success  = 0;
  static int DSN_off  = 0;

#define BOL 1
  static struct menu_item dsn_items[] = {
      { "Return H)eaders only on FAILURE:", 'h', 3,  BOL, 
	{ (char *) &hdr_only},
	sizeof hdr_only },
      { "Return DSN on F)AILURE         :", 'f', 5,  BOL, 
	{ (char *) &failure },
	sizeof failure },
      { "              D)ELAY           :", 'd', 6,  BOL, 
	{ (char *) &delay },
	sizeof delay },
      { "              S)UCCESS         :", 's', 7,  BOL, 
	{ (char *) &success },
	sizeof success },
      { "U)se defaults (DSN off)        :", 'u', 12, BOL, 
	{ (char *) &DSN_off },
	sizeof DSN_off }
  };

  if (0 == *dsn)
      DSN_off = 1;
  else {
    if (DSN_FULL    & *dsn) hdr_only = 0;
    if (DSN_HDRS    & *dsn) hdr_only = 1;
    if (DSN_SUCCESS & *dsn) success  = 1;
    if (DSN_FAILURE & *dsn) failure  = 1;
    if (DSN_DELAY   & *dsn) delay    = 1;
    if (DSN_NEVER   & *dsn) { success = 0; failure = 0; delay = 0; }
  }
  
  if (generic_menu(dsn_items,sizeof dsn_items / sizeof (struct menu_item),
	       "DSN (Delivery Status Notification) Configuration",
		   "DSN: ")) {

      if (DSN_off) *dsn = 0;
      else if (!success && !failure && !delay) *dsn = DSN_NEVER;
      else {
	  int val = 0;
	  if (hdr_only) val |= DSN_HDRS;
	  else          val |= DSN_FULL;
	  if (success)  val |= DSN_SUCCESS;
	  if (failure)  val |= DSN_FAILURE;
	  if (delay)    val |= DSN_DELAY;
	  
	  *dsn = val;
      }
  }
}

static int 	    edit_return_path(struct mailing_headers *headers,
				     struct menu_context *page);
static int 	    edit_return_path(headers, page)
     struct mailing_headers *headers;
     struct menu_context *page;
{

    if (headers->env_from) {
	int ret;
	int can_edit;
	const char *t = mailer_env_from_value(headers->env_from,&can_edit);
	struct string *buffer = NULL;

	int LINES, COLUMNS;

	menu_get_sizes(page,&LINES, &COLUMNS);

	if (!can_edit)
	    return 0; /* bad command */

	MoveCursor(LINES-4,0);
	CleartoEOS();
	       
	print_format_center(LINES-4, 
			    CATGETS(elm_msg_cat, ElmSet,
				    ElmEnterEnvFrom, 
				    "Enter value for envelope sender address."));

	if (t)
	    buffer = new_string2(display_charset,cs2cus(t));

	/* NOTE: Currently alias expansion is not possble */

	ret = optionally_enter2(page,
				&buffer,LINES-2, 0, 
				OE_REDRAW_MARK|
				OE_SIG_CHAR /* Ctrl-C */,
				CATGETS(elm_msg_cat, ElmSet, 
					ElmPreSEnvFromPrompt,
					"Return-path: "));
	
	if (ret == 0 && buffer) {
	    char * s = us2s(stream_from_string(buffer,0,NULL));

	    mailer_env_from_change(headers->env_from,s);

	    free(s);
	}
	
	if (buffer)
	    free_string(&buffer);
	
	if (ret == REDRAW_MARK)
	    return REDRAW_MARK;

	if (0 != ret)       /* Error or Ctrl-C */
	    return 1;

	ClearLine(5);

	t = mailer_env_from_value(headers->env_from,&can_edit);

	if (can_edit) {
	    if (t)
		PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFrom,
				     "R)eturn-path: %s"),
			 t);
	    else
		PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromDef,
				     "R)eturn-path: (default)"));
	} else if (t)
	    PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromNoEdit,
				 "Return-path: %s"),
		     t);

	return 1;
    }
    return 0; /* bad command */
}

/*
 * verify_transmission() - Ask the user to confirm transmission of the
 * message.  Returns 0 to send it, -1 to forget it.
 *                                 -2 to forget empty file
 */
static int verify_transmission(filename, 
			       already_has_text, copy_file, force_cmd, options,
			       dsn, headers, attachments,
			       file_set,
			       mailer_info,
			       hdr_charset, mailbox,
			       encode_hdr,
			       aview,page,
			       prompt_area,
			       parent_message)
     char *filename;	    /* pathname to mail mssg composition file        */
     int already_has_text;  /* indicates there is already text in the mssg   */
     struct copy_file *copy_file; /* pointer to buffer holding copy file name   */
     int force_cmd;	    /* command to do, '\0' to prompt user for cmd    */
     int options;
     int *dsn;
     struct mailing_headers *headers;   
     struct Attachments *attachments;
     charset_t file_set;      /* charset of file (filename) */
     struct mailer_info  *mailer_info;
     charset_t hdr_charset;
     struct MailboxView *mailbox;   /* needed for system_call() */
     int encode_hdr;
     struct AliasView *aview;
     struct menu_context *page;    /* new menu context assumed! */
     struct menu_context  *prompt_area;
     struct header_rec *parent_message;   /* Replying .... */
{
    char *prompt_mssg = NULL;	/* message to display prompting for cmd	*/
    char prompt_menu1[SLEN],	/* menu of available commands		*/
	prompt_menu2[SLEN];
    int bad_cmd;		/* set TRUE to bitch about user's entry	*/
    int did_prompt;		/* TRUE if cmd prompted for and entered	*/
    int cmd;			/* command to perform			*/
    char lbuf[VERY_LONG_STRING];
    int x_coord, y_coord;
    int replying          = 0 != (options & MAIL_REPLYING);
    int have_attachments  = 0 != (options & MAIL_HAVE_ATTACHMENTS);
#ifdef USE_PGP
    int was_pgp_encoded   = 0 != (options & MAIL_HAVE_PGP_ENCODED);
#endif
    int def_cmd = 'y';
    int do_redraw = 1;
    int LINES, COLUMNS;
    int ul = give_dt_enumerate_as_int(&user_level);
    
    int ret = -1;
    int menu_build = 0;

    menu_get_sizes(page,&LINES, &COLUMNS);

#ifdef USE_PGP
    if (replying && was_pgp_encoded &&
	! (pgp_status & PGP_MESSAGE)) 
	def_cmd = 'P';
#endif

    bad_cmd = FALSE;		/* nothing to complain about yet	*/

    for (;;) {
	menu_set_default(page);

	if (0) {
	redraw:
	    do_redraw = 1;
	}

    resize_mark:
	if (menu_resized(page)) {
	    
	    menu_get_sizes(page,&LINES, &COLUMNS);

	    menu_subpage_relocate(prompt_area,page,LINES-4,4);

	    do_redraw = 1;

	} else if (menu_need_redraw(page))
	    do_redraw = 1;
	
	if (do_redraw) {

	    do_redraw = 0;
	    
	    menu_ClearScreen(page);
	    menu_StartXX(page,pg_BOLD);

	    menu_print_format_center(page,0, 
				     CATGETS(elm_msg_cat, ElmSet,
					     ElmMailScreenTitle, 
					     "Mail Pre-Send Screen"));

	    menu_EndXX(page,pg_BOLD);

	    show_last_error();

	    show_presend_headers(headers,hdr_charset, page);

	    if (headers->env_from && ul > 0) {
		int can_edit;
		const char *t = mailer_env_from_value(headers->env_from,&can_edit);

		if (can_edit) {
		    if (t)
			PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFrom,
					     "R)eturn-path: %s"),
				 t);
		    else
			PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromDef,
					     "R)eturn-path: (default)"));
		} else if (t)
		    PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromNoEdit,
					 "Return-path: %s"),
			     t);				     
	    }

	    menu_redraw_children(page);

	}

	/* build up prompt and menu strings */
	if (menu_build) {
	    ; /* not changed - no need to rebuild the strings */
	} else if (ul == 0) {
	    prompt_mssg = catgets(elm_msg_cat, ElmSet,
				  ElmVfyPromptSendTheMsg,
				  "Send the message now? y");

	    strfcpy(prompt_menu1, catgets(elm_msg_cat, ElmSet,
					  ElmVfyMenu1User0,
					  "Select letter of header to edit, 'e' to edit the message,"),
		    sizeof prompt_menu1);   
	    strfcpy(prompt_menu2, catgets(elm_msg_cat, ElmSet,
					 ElmVfyMenu2User0,
					 "'a' to make attachments, 'y' to send message, or 'n' to cancel."),
		    sizeof prompt_menu2);   

	    /* In some conditions add these also to menu in user_level == 0 */

	    if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		DPRINT(Debug,10,(&Debug,"mailer supports DSN\n"));
		
		if (dsn && *dsn != 0) {
		    strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
						  ElmVfyMenuDSN,
						  "D)SN"), 
			    sizeof prompt_menu2);
		}
	    }

#ifdef USE_PGP
	    if (replying && was_pgp_encoded) {
		if (prompt_menu2[0]) strfcat(prompt_menu2,", ",
					     sizeof prompt_menu2);
		strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
					      ElmVfyMenuPGP,
					      "P)gp"), 
			sizeof prompt_menu2);
	    }
#endif
	} else {

	    prompt_mssg = catgets(elm_msg_cat, ElmSet,
				  ElmVfyPromptSendMessage,
				  "Send message now? y");
	    strfcpy(prompt_menu1, catgets(elm_msg_cat, ElmSet,
					  ElmVfySelectLetter,
					  "Select letter of header to edit, "),
		    sizeof prompt_menu1);


	    strfcat(prompt_menu1, catgets(elm_msg_cat, ElmSet,
					  ElmVfyMenuEditMsg, 
					  "e)dit message, "),
		    sizeof prompt_menu1);


	    strfcat(prompt_menu1, catgets(elm_msg_cat, ElmSet,
					  ElmVfyMenuHeaders1, 
					  "all h)eaders, "),
		    sizeof prompt_menu1);

	    
	    strfcpy(prompt_menu2,catgets(elm_msg_cat, ElmSet,
					ElmVfyMenuVfyCpy,
					"a)ttachments, co(p)y, "), 
		    sizeof prompt_menu2);
#ifdef ISPELL
	    strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet, ElmVfyMenuIspell,
					  "i)spell, "),
		    sizeof prompt_menu2);
#endif


#ifdef ALLOW_SUBSHELL
	    strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet, ElmVfyMenuShell,
					 "!)shell, "),
		    sizeof prompt_menu2);
#endif


	    if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		DPRINT(Debug,10,(&Debug,"mailer supports DSN\n"));
		
		if (dsn)
		    strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
						  ElmVfyMenuDSN,
						  "D)SN, "), 
			    sizeof prompt_menu2);
	    }


#ifdef USE_PGP
	    strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
					  ElmVfyMenuPGP,
					  "P)gp, "), 
		    sizeof prompt_menu2);
#endif

	    if (mailbox) {
		strfcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
					      ElmVfyMenuFolder,
					      "view) folder, "), 
			sizeof prompt_menu2);
	    }

	    strcat(prompt_menu2, catgets(elm_msg_cat, ElmSet,
					 ElmVfyMenuForget1,
					 "or f)orget"));
	   
    	    menu_build = 1;

	}

	/* complain if last entry was bad */
	if (bad_cmd) {
	    menu_Write_to_screen(prompt_area,
				 FRM("%c??"), 07);
	    if (sleepmsg > 0)
		error_sleep((sleepmsg + 1) / 2);
	    bad_cmd = FALSE;
	}

	if (menu_resized(prompt_area) ||
	    menu_need_redraw(prompt_area)) {
	
	    menu_ClearScreen(prompt_area);
	 
	    menu_PutLine0(prompt_area,
			  0, 0, prompt_mssg);

	    Centerline(1, prompt_menu1,prompt_area);
            Centerline(2, prompt_menu2,prompt_area);

	    show_last_error();
	}

	/* if we don't have a cmd, display prompt and get response from user */
	if (force_cmd != '\0' &&
	    isascii(force_cmd)) {
	    cmd = tolower(force_cmd);
	    force_cmd = '\0';
	    did_prompt = FALSE;
	} else {
	    menu_PutLine0(prompt_area,
			  0, 0, prompt_mssg);

	    menu_GetXYLocation(prompt_area,
			       &x_coord, &y_coord);
	    y_coord--; /* backspace over default answer */

	    menu_CleartoEOLN(prompt_area);

	    menu_PutLineX(prompt_area,
			  0, y_coord, FRM("%c"),def_cmd);
	    FlushBuffer();

	    menu_MoveCursor(prompt_area,
			    x_coord, y_coord);
	    cmd = menu_ReadCh(prompt_area,REDRAW_MARK|READCH_resize);

	    if (cmd == REDRAW_MARK) {
		menu_ClearScreen(page);   /* Clear possible redraw mark */
		
		/* Call refresh routines of children */
		menu_redraw_children(page);

		if (menu_need_redraw(prompt_area))		    
		    menu_ClearScreen(prompt_area);   /* Clear redraw mark from prompt_area*/

		goto redraw;
	    }

	    if (cmd == RESIZE_MARK) {

		DPRINT(Debug,4, (&Debug,"       .... resizing\n"));
		goto resize_mark;
	    }

	    if (cmd == EOF)
		goto handle_EOF;

	    did_prompt = TRUE;
	    clear_error();

	    /* Put cursor pack to correct place */
	    menu_MoveCursor(prompt_area,
			    x_coord, y_coord);
	}

	switch (cmd) {
	case '\n':
	case '\r':
	    cmd = def_cmd;
	}

	/* handle command */
	switch (cmd) {
	case '?':
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     FRM("?"));
	    
	    {
		struct elm_commands *cmds = give_send_commands();
		
		int ch = help_generic(cmds,TRUE,page,prompt_area);
		
		free_commands(&cmds);
		
		if (menu_need_redraw(page)) {
		    menu_ClearScreen(page);
		    
		    do_redraw = 1;
		}
		
		if (EOF == ch) {
		    ret = -3;
		    goto handle_EOF;
		}		    		    
	    }
	    break;

	case 'a':
	    if (did_prompt)
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuAtttach,
					     "Attachments"));

	    attach_menu(NULL,NULL,attachments,display_charset,mailer_info,
			NULL, aview,mailbox,NULL);
	    do_redraw = 1;
	    break;

        case 'D':
	    if (dsn) {
		
		if (did_prompt)
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuDsn,
						 "Dsn"));
		
		
		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		    DPRINT(Debug,8,(&Debug,"mailer supports DSN\n"));
		    
		    dsn_menu(dsn);
		    do_redraw = 1;
		} else {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmVfyMenuDsnNotSupported,
				      "Mailer does not support delivery notification settings."));
		    
		}
	    } 
	    break;

	case ctrl('D') :
	    if (dsn) {

		if (did_prompt)
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuDsnToggle,
						 "Dsn toggle"));
		

		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {

		    if (0 == *dsn || 
			0 != (*dsn & DSN_NEVER) ||
			0 == (*dsn & DSN_SUCCESS)) {

			*dsn &=  ~DSN_NEVER;
			*dsn |=   DSN_SUCCESS|DSN_FAILURE|DSN_DELAY;
		
			lib_transient(CATGETS(elm_msg_cat, ElmSet, 
					      ElmVfyMenuDsnEnabled,
					      "Delivery notifications enabled.")); 

		    } else if (0 != (*dsn & DSN_SUCCESS)) {
			*dsn &=   ~DSN_SUCCESS;

			if (0 == *dsn) {
			    *dsn = DSN_NEVER;
			    lib_transient(CATGETS(elm_msg_cat, ElmSet, 
						  ElmVfyMenuDsnDisabled,
						  "Delivery notifications disabled."));
			
			} else
			    lib_transient(CATGETS(elm_msg_cat, ElmSet, 
						  ElmVfyMenuDsnSucceesDisabled,
						  "Delivery notification for SUCCESS disabled."));
			
		    } 
		    
		} else {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmVfyMenuDsnNotSupported,
				      "Mailer does not support Dsn notification settings."));
		    
		}		    
	    }
	    break;

	case 'y':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuSend,
					     "Send"));
		/* Call FlushBuffer() after writing of "Send" so that
		 * user get immediately feedback (it may take time
		 * to encode big attachments)
		 */
		FlushBuffer();
	    }
#ifdef USE_PGP
	      if (replying && was_pgp_encoded &&
		  ! (pgp_status & PGP_MESSAGE)) {

		  menu_ClearLine(prompt_area,1);
		  menu_PutLineX(prompt_area,1, 0, 
				CATGETS(elm_msg_cat, ElmSet, ElmPgpRcvSure, 
					"The recv'd message was PGP encoded, are you sure? "));
		  for (;;) {
		      cmd = menu_ReadCh(prompt_area,0);
		      if (cmd == *def_ans_yes) {
			  menu_Write_to_screen(prompt_area,
					       CATGETS(elm_msg_cat, ElmSet, ElmYesWord, 
						       "Yes."));
			  ret = 0;
			  goto out;
		      }
		      if (cmd == *def_ans_no) {
			  menu_Write_to_screen(prompt_area,
					       CATGETS(elm_msg_cat, ElmSet, ElmNoWord, 
						       "No."));
			  break;
		      }
		  }
		  break;
	      }
#endif /* USE_PGP */
	      ret = 0;
	      goto out;
	      /*NOTREACHED*/

	case ctrl('F'):
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet,
					     ElmPassphraseForget,
					     "Forget passphrase"));
	    forget_passphrase();	      
	    break;

	case 'f':
	case 'n':
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuForget, 
					     "Forget"));
#ifdef USE_PGP
            pgp_status = 0; /* make sure to reset! */
#endif
	    if (bytes(filename) <= 0) {
		; /* forget about empty files */
		ret = -2;
		goto out;
	    } 
	    ret =  -1;
	    goto out;
	    /*NOTREACHED*/

	case 'p':
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuCopyFile, 
					     "Copy file"));
	    name_copy_file(copy_file, aview, mailbox,page);
	    menu_trigger_redraw(prompt_area);   
	    break;

	case 'e': 
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuEdit, "Edit"));

	    {
		char * editor_val = give_dt_estr_as_str(&editor_e,"editor");

		if (editor_val) {
		    do_redraw = 1;
		    if (edit_the_message(filename, already_has_text,
					 headers,editor_val,file_set,
					 mailer_info, mailbox,
					 aview, page) != 0) {
			ret =  -1;
			goto out;
		    }
		} else {
		    sleep_message();
		    if (no_editor_edit_the_message(filename,headers,file_set,
						   mailer_info,mailbox,aview)) {
			ret =  -1;
			goto out;
		    }
		}
	    }
	    break;

	case 'h':
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuHeaders,"Headers"));
	    edit_headers(headers,mailer_info,hdr_charset,encode_hdr,
			 aview,parent_message);

	    

	    do_redraw = 1;
	    break;

	case 'm':     /* make form ---  AT&T Mail Forms support is dropped */
	    bad_cmd = TRUE;
	    break;

#ifdef ISPELL
	case 'i':
	    if (did_prompt)
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuIspell2,"Ispell"));


		elm_sfprintf(lbuf, sizeof lbuf,
			     FRM("%s %s %s"),
			     ISPELL_PATH, ISPELL_OPTIONS, filename);
		Raw(OFF);
		system_call(lbuf, SY_ENAB_SIGHUP, NULL);

		do_redraw = 1;
		Raw(ON);
	    break;
#endif

#ifdef ALLOW_SUBSHELL
	case '!':
	    menu_ClearLine(prompt_area,0);
	    subshell(NULL, page, prompt_area);

	    if (menu_need_redraw(page)) {
		menu_ClearScreen(page);

		do_redraw = 1;
	    }
	    break;
#endif
#ifdef USE_PGP
        case 'P':
	    if (did_prompt)
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuPgp,"Pgp"));
	    if (!pgp_status) {
	      pgp_status = pgp_menu (filename,headers, page);
	      if (pgp_status)
		  def_cmd = 'y';
	      else
		  def_cmd = 'P';
	      do_redraw = 1;
	    }
            else
	      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmPgpAlreadyEncSig,
				"This message is already encrypted and/or signed!"));
            break;
#endif
	case 'w':
	    if (mailbox) {
		int ch;

		if (did_prompt)
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, 
						 ElmVfyMenuViewFolder,
						 "View folder"));

		ch = send_view_folder(mailbox, attachments, page, aview,
				 mailer_info);
		
		if (menu_need_redraw(page)) {
		    menu_ClearScreen(page);

		    do_redraw = 1;
		}

		if (EOF == ch) 
		    goto handle_EOF;

	    } else
		bad_cmd = TRUE;
	    break;
    
	case 'R':
	    {
		int c = edit_return_path(headers, page);

		if (0 == c)
		    bad_cmd = TRUE;
		else if (REDRAW_MARK == c)
		    goto redraw;		    

		menu_trigger_redraw(prompt_area);
	    }
	    break;
	default:

	    {
		int c = presend_action(headers,mailer_info,cmd,hdr_charset,
				       aview, page, prompt_area,
				       parent_message);

		if (0 == c) {   /* Accept also uppercase letters */
		    int old = cmd;
#ifdef ASCII_CTYPE
		    if (isascii(cmd))
#endif
			cmd = tolower((unsigned char)cmd);
		    
		    if (old != cmd)
			c = presend_action(headers,mailer_info,cmd,hdr_charset,
					   aview, page, prompt_area, parent_message);

		    if (0 == c)
			bad_cmd = TRUE;
		    else if (REDRAW_MARK == c)
			goto redraw;		    
		    
		} else if (REDRAW_MARK == c)
		    goto redraw;		 

		menu_trigger_redraw(prompt_area);   
	    }
	    
	    break;
	}
    }

    if (0) {
	/* TODO: Propagate EOF */
	
    handle_EOF:
	DPRINT(Debug,2,(&Debug,"EOF on verify_transmission\n"));
	
	leave(0);   
	ret = -3;
    }

 out:

    return ret;
}

void write_header_info(mailer, headers,
		       copy, mime_info)
     struct out_state *mailer;
     struct mailing_headers * headers;  
     int   copy;
     mime_send_t *mime_info;
{
    /**
       
        Added the ability
	to have backquoted stuff in the users .elmheaders file!
	If copy is TRUE, then treat this as the saved copy of outbound
	mail.
    **/
    

    mime_info -> encode_hdr = !allow_no_hdrencoding;
	
    /** Subject moved to top of headers for mail because the
	pure System V.3 mailer, in its infinite wisdom, now
	assumes that anything the user sends is part of the 
	message body unless either:
	1. the "-s" flag is used (although it doesn't seem
	   to be supported on all implementations? )
	2. the first line is "Subject:".  If so, then it'll
	   read until a blank line and assume all are meant
	   to be headers.
	So the gory solution here is to move the Subject: line
	up to the top.  I assume it won't break anyone elses program
	or anything anyway (besides, RFC-822 specifies that the *order*
	of headers is irrelevant).  Gahhhhh....
    **/

    if (headers->subject)
	write_string_header(mailer,"Subject",
			    headers->subject,
			    mime_info->encoding_top,
			    mime_info->encode_hdr,
			    mime_info->hdr_charset);

    {
	const char *MIME_name_h = 
	    get_charset_MIME_name(mime_info->hdr_charset);


	/* Our standard violations */
	char OSV[100];
	OSV[0] = '\0';
	  
	if (!mime_info->encode_hdr && MIME_name_h)
	    add_parameter(OSV,"hdr-charset",
			  MIME_name_h,sizeof(OSV),FALSE);
	
	/* If allow_no_hdrencoding is set, there is no mime-encoding on
	   user headers either
	*/
	if (allow_no_hdrencoding) {
	    add_parameter(OSV,"no-hdr-encoding","1",sizeof(OSV),FALSE);	    
	}

	if (OSV[0]) {
	    state_printf(mailer, 
			 FRM("X-ELM-OSV: (Our standard violations) %s"),
			 OSV);
	    print_EOLN(mailer,mime_info->encoding_top);
	}
    } 
    
    if (headers->in_reply_to)
	write_references_header(mailer,"In-Reply-To",
				headers->in_reply_to,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);
    if (headers->references)
	write_references_header(mailer,"References",
				headers->references,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);


    if (headers->to.addrs)
	write_addr_header(mailer,"To",headers->to.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
    write_text_header(mailer,"Date",get_arpa_date(),
		      mime_info->encoding_top);

    if (headers->sender)
	write_text_header(mailer,"Sender",headers->sender,
			  mime_info->encoding_top);
	
    if (headers->from.addrs) 
	write_addr_header(mailer,"From",headers->from.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,0);
    
    if (headers->cc.addrs)
	write_addr_header(mailer,"CC",headers->cc.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);

    if ((copy && headers->bcc.addrs) ||
	/* If there is no To or CC recipients,
	 * show BCC recipients to make mail legal 
	 */
	(!headers->to.addrs && !headers->cc.addrs && headers->bcc.addrs))
	write_addr_header(mailer,"Bcc",headers->bcc.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
	
    if (headers->action)
	write_text_header(mailer,"Action",headers->action,
			  mime_info->encoding_top);

    if (headers->priority)
	write_text_header(mailer,"Priority",headers->priority,
			  mime_info->encoding_top);

    if (headers->precedence)
	write_text_header(mailer,"Precedence",headers->precedence,
			  mime_info->encoding_top);

    if (headers->expires)
	write_text_header(mailer,"Expires",headers->expires,
			  mime_info->encoding_top);

    if (headers->reply_to.addrs)
	write_addr_header(mailer,"Reply-To",headers->reply_to.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
	
    if (headers->user_header_count > 0) {
	int i;
	for (i = 0; i <  headers->user_header_count; i++) {
	    write_string_header(mailer,
				give_header_name(headers->user_header[i].
						 name),
				headers->user_header[i].value,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);
	}
    }

#ifndef NO_XHEADER
    state_printf(mailer, FRM("X-Mailer: ELM [version %s]"), version_buff);
    print_EOLN(mailer,mime_info->encoding_top);
#endif /* !NO_XHEADER */

    mime_write_top_headers (mailer, mime_info);

    mime_info->cl_offset = 0;
    if (copy) {
	const char * t = NULL;
	int x;

	if (headers->env_from)
	    t = mailer_env_from_value(headers->env_from,&x);

	if (t)
	    write_text_header(mailer,"Return-path",t,
			      mime_info->encoding_top);
	
	/* Only write content-length to copy */
	state_puts("Content-Length: ",mailer);
	mime_info->cl_offset = out_state_ftell(mailer);
	state_puts("          ",mailer); /* Print Blanks as Placeholders */
	print_EOLN(mailer,mime_info->encoding_top);
	state_puts("Status: RO",mailer);
	print_EOLN(mailer,mime_info->encoding_top);
    }
    print_EOLN(mailer,mime_info->encoding_top);
}

static int copy_part P_((mime_send_t *mime_info,struct out_state *dest,
			 FILE *conv_file,struct mime_send_part * X,
			 int copy));
static int copy_part(mime_info,dest,conv_file,X,copy)
     mime_send_t *mime_info;
     struct out_state *dest;
     FILE *conv_file;
     struct mime_send_part * X;
     int copy;
{
    int line_len;
    char buffer[SLEN];			/* file reading buffer */
    long a = out_state_ftell(dest);
    long pos1;

    if (0 != fseek(conv_file,X->start_loc,SEEK_SET)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSeekFailed,
			  "ELM [seek] failed trying to read %d bytes into file."),
		  (int) X->start_loc);
	return 0;
    }

       
    while ((pos1 = ftell(conv_file)) < X->end_loc &&
	   (line_len = mail_gets(buffer, sizeof buffer -1, conv_file))) {

	/* Text part necessary do not end to \n so there is possibility to
	   get part of next part 
	*/

	if (pos1 + line_len > X->end_loc) {
	    int old = line_len;

	    line_len = X->end_loc - pos1;

	    DPRINT(Debug,4,(&Debug,"copy_part: Got past of end of attachment -- changing line_len %d -> %d\n",
			    old,line_len));
	    if (line_len < 0 || line_len > old)
		mime_panic(__FILE__,__LINE__,
			   "copy_part",
			   "line_len calculation error");

	    if (0 == line_len)
		break;

	}


	if (X->encoding_part == ENCODING_BINARY) {
	    /* \n -> \r\n conversion is done for that part already */
	    state_put(buffer, line_len, dest);
	} else if (mime_info->encoding_top == ENCODING_BINARY) {
	    /* It is better perhaps use canonical eol (CRLF) when mail have
	     * content transfer encoding BINARY somewhere (see notes about 
	     * BINARYMIME)
	     */
	    
	    /* THIS assumes that 
	       dest->EOLN_is_CRLF is set
	    */
	    state_convert_EOLN(buffer,&line_len,sizeof buffer,dest);

	    /* We don't do escaping because line does not end either 
	     * to plain \n now...
	     */
	    state_put(buffer, line_len, dest);
#ifndef DONT_ESCAPE_MESSAGES
	} else if (copy && (strncmp(buffer, "From ", 5) == 0)) {
	    /* Add in the > to a From on our copy */
	    state_puts(">",dest);
	    state_put(buffer, line_len, dest);
#endif
#ifdef NEED_LONE_PERIOD_ESCAPE
	} else if (!copy && strcmp(buffer, ".\n") == 0) {
	    /* Because some mail transport agents take a lone period to
	     * mean EOF, we add a blank space on outbound message.
	     */
	    state_puts(". \n", dest);
#endif /* NEED_LONE_PERIOD_ESCAPE */
	} else {
	    /* SMTP uses \r\n */
	    if (out_state_EOLN_is_CRLF(dest))
		state_convert_EOLN(buffer,&line_len,sizeof buffer,dest);

	    state_put(buffer, line_len, dest);
	}
    }

    if (out_state_ferror(dest) || ferror(conv_file))
	return 0;

    DPRINT(Debug,4, (&Debug,
		     "Preparing mail for %s: part %d -- %d bytes written to temp%s\n",
		     copy ? "copy" : "sending",
		     X-mime_info->top_parts,
		     out_state_ftell(dest)-a,
		     feof(conv_file) ? ", got EOF" : ""));

    return 1;
}

int copy_message_across(mime_info, dest, copy, conv_file)
     mime_send_t *mime_info;
     struct out_state *dest;
     int copy;
     FILE *conv_file;
{
    /** If copy is TRUE, treat as a saved copy of outbound mail. **/

    mime_info -> cl_start = out_state_ftell(dest);

    if (mime_info->msg_is_multipart) {
	int i;

	DPRINT(Debug,4, (&Debug,
			 "Preparing mail for %s: %d parts (multipart)\n",
			 copy ? "copy" : "sending",
			 mime_info->top_parts_count));

	for (i = 0; i < mime_info->top_parts_count; i++) {
	    if (!copy || mime_info->top_parts[i].save_it_on_copy) {

		print_EOLN(dest,mime_info->encoding_top);
		state_printf(dest,FRM("--%s"),mime_info->mime_boundary);
		print_EOLN(dest,mime_info->encoding_top);

		mime_write_part_headers(dest,mime_info,
					&(mime_info->top_parts[i]));
		print_EOLN(dest,mime_info->encoding_top);

		if (!copy_part(mime_info,dest,conv_file,
			       &(mime_info->top_parts[i]),copy)) {
		    return 0;
		}
	    }
	}
	print_EOLN(dest,mime_info->encoding_top);
	state_printf(dest,FRM("--%s--"), mime_info->mime_boundary);
	print_EOLN(dest,mime_info->encoding_top);

    } else if (1 == mime_info->top_parts_count) {
	DPRINT(Debug,4,(&Debug,  
		  "Preparing mail for %s: one part (not multipart)\n",
		  copy ? "copy" : "sending"));

	if (!copy || mime_info->top_parts[0].save_it_on_copy) {
	    if (!copy_part(mime_info,dest,conv_file,
			   &(mime_info->top_parts[0]),copy)) {
		return 0;
	    }
	}
    } else {
	DPRINT(Debug,4,(&Debug,  
			"Preparing mail for %s: ODD: %d parts (NOT multipart)\n",
			copy ? "copy" : "sending",mime_info->top_parts_count));
    }

    mime_info->cl_end = out_state_ftell(dest) ;
    if (mime_info->cl_offset > 0) {
	/* go fixup the content length header */
	out_state_fseek(dest, mime_info->cl_offset);
	state_printf(dest, FRM("%ld"), 
		     (long)(mime_info->cl_end - mime_info->cl_start));
	  
	/* Return to the end of the file! */
	out_state_fseek (dest, mime_info->cl_end);
    }

    DPRINT(Debug,4,(&Debug,  
		    "Preparing mail for %s: total %d bytes body, ~%d bytes headers\n",
		    copy ? "copy" : "sending",
		    mime_info->cl_end - mime_info -> cl_start,
		    mime_info -> cl_start));
    
    /* !! */
    if (EOF == fflush(out_state_FILE(dest)) || ferror(out_state_FILE(dest))) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmWriteFailedCopy,
			  "Write failed to temp file in copy"));
	return 0;
    }

    return 1;
}

static int have_remote_address P_((struct addr_list * addr_list));
static int have_remote_address(addr_list)
     struct addr_list * addr_list;
{
    int idx;
    int len = addr_list_item_count(addr_list);
    
    /* check To: list */
    
    for (idx = 0; idx < len; idx++) {
	
	int group = -1;
	const struct address * address = 
	    addr_list_get_item(addr_list,idx,&group);
	
	const char  * ad  = address_get_ascii_addr(address);
	
	const char *ptr;

	if (!ad)
	    continue;
	
	ptr = qstrpbrk_c(ad,":@");
	
	/* Have route, so address is remote */
	if (ptr && ':' == *ptr)
	    return 1;

	/* Address do not have '@' character, 
	   so this address is local         */
	if (!ptr)
	    continue;

	ptr++;
	
	if (istrcmp(ptr, hostfullname) != 0) {
	    return 1;
	}
    }

    return 0;
}

static int append_sig(file, headers, page)
     FILE *file;
     struct mailing_headers *headers;
     struct menu_context *page;
{
    /* Append the correct signature file to file.  Return TRUE if
       we append anything.  */
    
    /* Look at the to and cc list to determine which one to use */
    
    /* We could check the bcc list too, but we don't want people to
       know about bcc, even indirectly */
    
    /* Some people claim that  user@anything.same_domain should be 
       considered local.  Since it's not the same machine, better be 
       safe and use the remote sig (presumably it has more complete
       information).  You can't necessarily finger someone in the
       same domain. */

    char * lsig = give_dt_estr_as_str(&local_signature_e, "localsignature");
    char * rsig = give_dt_estr_as_str(&remote_signature_e,"remotesignature");


    if (!lsig || !rsig)
	error_wait();

    /* FIXME -- is this correct? */
    if (! OPMODE_IS_BATCHMODE(opmode) && (lsig || rsig)) {

	char filename2[SLEN];
	char *sig = "";

	/* check each @ for @thissite.domain */
	/* if any one is different than this, then use remote sig */

	sig = lsig;

	if (headers->to.addrs &&
	    have_remote_address(headers->to.addrs))
	    sig = rsig;	   	    
	

	/* still local? */         
	if (sig == lsig &&
	    headers->cc.addrs &&
	    have_remote_address(headers->cc.addrs))
	    sig = rsig;
    
	
	if (sig && sig[0]) {  /* if there is a signature file */
	    if (sig[0] != '/')
		elm_sfprintf(filename2, sizeof filename2,
			     FRM("%s/%s"), 
			     home, sig);
	    else
		strfcpy(filename2, sig, sizeof filename2);

	    /* append the file - with a news 2.11 compatible */
	    /* seperator if "sig_dashes" is enabled */
	    (void) append(file, filename2, (sig_dashes ? "\n-- \n" : NULL),
			  page);
	    
	    return TRUE;
	}    
    }
    return FALSE;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
